# +
#   ARES/HADES/BORG Package -- -- ./scripts/generate_tests_cmake.py
#   Copyright (C) 2019 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2019)
#
#   This program is free software; you can redistribute it and/or modify it
#   under the terms of either the CeCILL license or the GNU General Public
#   license, as included with the software package.
#
#   The text of the license is located in Licence_CeCILL_V2.1-en.txt
#   and GPL.txt in the root directory of the source package.
# +

import os
import sys
import filecmp

output_cmake_file = sys.argv[1]
input_config = sys.argv[2]
base_dir = sys.argv[3]
input_config_path = os.path.dirname(os.path.abspath(input_config))

with open(input_config, mode="r") as f:
    config = {}
    exec(f.read(), {}, config)
    config = config['tests']

tests = config['tests']

with open(output_cmake_file, "wt") as f:

    for test_name in tests:
        test = tests[test_name]
        base_code = test.get("base_code", "generic_gradient_test.cpp")
        includes = test['includes']
        likelihood = test['likelihood']
        info_code = test.get('info_code', '')
        model_code = ""

        f.write("""
add_executable(test_gradient_%(test_name)s %(base_dir)s/%(test_source)s)
target_link_libraries(test_gradient_%(test_name)s test_library_LSS LSS ${LIBS})
ares_add_test_targets(test_gradient_%(test_name)s)
target_include_directories(test_gradient_%(test_name)s PRIVATE %(input_config_path)s)
""" % dict(test_name=test_name,
           base_dir=base_dir,
           input_config_path=input_config_path,
           test_source="test_%s_gradient.cpp" % (test_name, )))

        includes_str = \
            "\n".join(
                map(lambda x: "#include \"%s\"" % (x,), includes)
            )

        downgrade = test.get('downgrade', 1)

        if 'model' in test:
            if isinstance(test['model'], list):
                model_string = ""
                args = test.get('model_args', ['comm, box,box, 0.001'] +
                                ['comm, box, box, 1'] *
                                (len(test['model']) - 1))
                extra_code = test.get('model_extra', [""] * len(test['model']))
                extra_code_prev = test.get('model_extra_prev', [""] * len(test['model']))
                for m, a, e, eprev in zip(test['model'], args, extra_code, extra_code_prev):
                    model_string += f"""
            {{
              {eprev}
              auto m = std::make_shared<{m}>({a});
              {e}
              chain->addModel(m);
            }}"""

                model_code = f"""
          auto makeModel(LibLSS::MPI_Communication * comm, LibLSS::MarkovState& state, LibLSS::BoxModel box, LibLSS::BoxModel box2) {{
            using namespace LibLSS;
            auto chain = std::make_shared<ChainForwardModel>(comm, box);
            {model_string}
            return chain;
          }}
          """
            else:
                args = test.get('model_args', 'comm, box, box, 0.001')
                extra_prev = test.get('model_extra_prev', "")
                extra = test.get('model_extra', "")
                model_code = f"""
                auto makeModel(LibLSS::MPI_Communication * comm, LibLSS::MarkovState& state, LibLSS::BoxModel box, LibLSS::BoxModel box2) {{
                  using namespace LibLSS;
                  {extra_prev}
                  auto m = std::make_shared<{test['model']}>({args});
                  {extra}
                  return m;
                }}
                """
        else:
            assert 'model_code' in test
            model_code = test['model_code']

        if 'data_setup' in test:
            data_setup = '#include "%s"\n#define DATA_SETUP %s\n' % (
                test['data_setup'])
        else:
            data_setup = ''

        newfile = os.path.join(base_dir,
                               "test_%s_gradient.cpp_tmp" % (test_name, ))
        finalfile = os.path.join(base_dir,
                                 "test_%s_gradient.cpp" % (test_name, ))

        with open(newfile, mode="wt") as f2:
            f2.write("""%(includes)s
#include "libLSS/samplers/rgen/hmc/hmc_density_sampler.hpp"
#include "libLSS/physics/likelihoods/base.hpp"
%(data_setup_header)s

namespace L = LibLSS::Likelihood;
using LibLSS::LikelihoodInfo;
using LibLSS::HMCDensitySampler;

static const int DOWNGRADE_DATA = %(downgrade)d;

HMCDensitySampler::Likelihood_t makeLikelihood(LikelihoodInfo& info) {
  %(additional_info_code)s
  return std::make_shared<%(likelihood)s>(info);
}

%(model_code)s

#include "%(base_code)s"
""" % dict(includes=includes_str,
            data_setup_header=data_setup,
            additional_info_code=info_code,
            likelihood=likelihood,
            model_code=model_code,
            downgrade=downgrade,
            base_code=base_code))

        try:
            os.stat(finalfile)
            if not filecmp.cmp(finalfile, newfile):
                os.rename(newfile, finalfile)
            else:
                os.unlink(newfile)
        except FileNotFoundError:
            os.rename(newfile, finalfile)

# ARES TAG: authors_num = 1
# ARES TAG: name(0) = Guilhem Lavaux
# ARES TAG: email(0) = guilhem.lavaux@iap.fr
# ARES TAG: year(0) = 2019



#+
#   ARES/HADES/BORG Package -- ./scripts/misc/convert_2m++.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
import numpy as np
from scipy import constants as sconst

LIGHT_SPEED = sconst.c/1000. #In km/s

catalog = np.load("2m++.npy")

with open("2MPP.txt", mode="w") as f:
    
    cond = (catalog['flag_vcmb']==1)*(catalog['flag_zoa']==0)*(catalog['best_velcmb'] > 100)
    for i,c in enumerate(catalog[cond]):
        M = c['K2MRS'] - 5*np.log10(c['best_velcmb']/100*1e5)
        zo = c['velcmb']/LIGHT_SPEED
        z = zo
        
        f.write(
            "%d %lg %lg %lg %lg %lg %lg\n" %
                (i, np.radians(c['ra']), np.radians(c['dec']), zo, c['K2MRS'], M, z)
            )
            



#+
#   ARES/HADES/BORG Package -- ./scripts/ares_tools/visu/__init__.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+



#+
#   ARES/HADES/BORG Package -- ./scripts/ares_tools/visu/vtktools.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
import numpy as np
from vtk.util.numpy_support import get_vtk_array_type, create_vtk_array, get_numpy_array_type

def numpy_scalar_to_vtk(np_array):
    """This function converts a numpy scalar array to a VTK array.
    
    Args:
      np_array (np.array): A numpy array
      
    Returns:
      vtk.vtkArray: An array of the closest possible type of numpy array. The array is deep
                    copied to avoid SEGV.
    
    """
    atype = get_vtk_array_type(np_array.dtype)    
    array = create_vtk_array(atype)
    array.SetNumberOfComponents(1)
    ntype = get_numpy_array_type(atype)
    adata = np.ravel(np_array).astype(ntype)
    array.SetVoidArray(adata, len(adata), 1)
    
    copy = array.NewInstance()
    copy.DeepCopy(array)
    return copy

def numpy_vector_to_vtk(np_array):
    """This function converts a numpy scalar array to a VTK array.
    
    Args:
      np_array (np.array): A numpy array
      
    Returns:
      vtk.vtkArray: An array of the closest possible type of numpy array. The array is deep
                    copied to avoid SEGV.
    
    """
    if np_array.shape[3] != 3:
      raise ValueError()
      
    atype = get_vtk_array_type(np_array.dtype)    
    array = create_vtk_array(atype)
    array.SetNumberOfComponents(3)
    ntype = get_numpy_array_type(atype)
    adata = np.ravel(np_array).astype(ntype)
    array.SetVoidArray(adata, len(adata), 1)
    
    copy = array.NewInstance()
    copy.DeepCopy(array)
    return copy


def smooth_array(a, L=[1.0,1.0,1.0], R=0.1):
    a_hat = np.fft.rfftn(a)
    ik = [np.fft.fftfreq(iN, d=iL/iN)*2*np.pi for iN,iL in zip(a.shape,L)]
    k2 = ik[0][:,None,None]**2 + ik[1][None,:,None]**2 + ik[2][None,None,:a.shape[2]/2+1]**2
    
    a_hat *= np.exp(-0.5*k2*R**2)
    return np.fft.irfftn(a_hat)

def displacement_array(a, L=[1.0,1.0,1.0], R=0.1):
    a_hat = np.fft.rfftn(a)
    ik = [np.fft.fftfreq(iN, d=iL/iN)*2*np.pi for iN,iL in zip(a.shape,L)]
    k2 = ik[0][:,None,None]**2 + ik[1][None,:,None]**2 + ik[2][None,None,:a.shape[2]/2+1]**2
    
    b = np.empty(a.shape + (3,), dtype=np.float32)
    
    b_hat = -a_hat * 1j*ik[0][:,None,None]/k2
    b_hat[0,0,0]=0
    b[...,0] = np.fft.irfftn(b_hat)

    b_hat = -a_hat * 1j*ik[1][None,:,None]/k2
    b_hat[0,0,0]=0
    b[...,1] = np.fft.irfftn(b_hat)

    b_hat = -a_hat * 1j*ik[2][None,None,:a.shape[2]/2+1]/k2
    b_hat[0,0,0]=0
    b[...,2] = np.fft.irfftn(b_hat)
    
    return b


def setupImageData3D(img_data, np_array, dims=[1.0,1.0,1.0], name="numpy array"):
    """This function setups a 3D image data object.
    
    """
    shape = np_array.shape
    
    dx = (d0/(N0-1) for d0,N0 in zip(dims,shape))
    img_data.SetOrigin(*(-d0/2 for d0 in dims)) # default values
    img_data.SetSpacing(*dx)
    img_data.SetDimensions(*shape) # number of points in each direction
    
    array = numpy_scalar_to_vtk(np_array)
    
    img_data.GetPointData().AddArray(array)
    array.SetName(name)
    
    return array
    




#+
#   ARES/HADES/BORG Package -- ./extra/python/python/doc/convert.py
#   Copyright (C) 2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+

# This tool transforms the doc rst in python/doc to a bunch
# of hpp header file. They are used to generate the adequate docstring
# for each module/class/function bound with pybind11 from C++ to Python
#
# The text must follow the Google convention for docstrings, formatted with
# RestructuredText markers. In addition to those markers this tool recognize
# the following formatting:
#
# A class/module doc is marked with a line starting by "@class"
# A comment line is a line starting with "@@"
# A function/member "F" is documented by writing the line "@funcname:F" followed
# by its documentation
#

import sys
import os

TAG_FUNCNAME = "@funcname:"
TAG_CLASSNAME = "@class:"
TAG_CLASS = "@class"
TAG_COMMENT = "@@"


def convert_doc(target_dir, fname):

    _, fname_base = os.path.split(fname)
    fname_header, ext = os.path.splitext(fname_base)
    base_prefix = prefix = fname_header.replace(".", "_")
    fname_header = os.path.join(target_dir, fname_header + ".hpp")
    funcname = None
    block_started = False
    funcs = {}

    with open(fname, mode="rt") as ff, open(fname_header, mode="wt") as gg:

        def finish_block():
            if block_started:
                gg.write(")str\";\n\n")

        def begin_block(funcname):
            nonlocal block_started
            gg.write(f"static const char * __pydocstring_{funcname} = R\"str(")
            block_started = True

        for l in ff:
            if l.startswith(TAG_COMMENT):
                continue
            if l.startswith(TAG_FUNCNAME):
                finish_block()
                l.strip()
                funcname = l[len(TAG_FUNCNAME):]
                funcname = prefix + "_" + funcname
                if funcname in funcs:
                    print(f"{funcname} is already defined")
                    sys.exit(1)
                begin_block(funcname)
                funcs[funcname] = True
                continue

            # Order matter here
            if l.startswith(TAG_CLASSNAME):
                prefix = base_prefix + "_" + l[len(TAG_CLASSNAME):].strip()
                finish_block()
                begin_block(prefix)
                continue

            if l.startswith(TAG_CLASS):
                prefix = base_prefix
                finish_block()
                begin_block(prefix)
                continue
            if block_started:
                gg.write(l)

        finish_block()


if __name__ == "__main__":
    target_dir = sys.argv[1]
    for a in sys.argv[2:]:
        convert_doc(target_dir, a)

# ARES TAG: authors_num = 1
# ARES TAG: name(0) = Guilhem Lavaux
# ARES TAG: email(0) = guilhem.lavaux@iap.fr
# ARES TAG: year(0) = 2020



#+
#   ARES/HADES/BORG Package -- ./scripts/misc/check_velocities.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
from read_all_h5 import *
import pylab as plt

g=read_all_h5('dump_velocities.h5')


V = g.scalars.L0[0]*g.scalars.L1[0]*g.scalars.L2[0]

q = g.scalars.k_pos_test
H=100.
D=1.
a=1.
f=g.scalars.cosmology['omega_m']**(5./9)
vref = 2* q/((q**2).sum()) / V * g.scalars.A_k_test * f * H * a**2 * D
vborg = g.scalars.lpt_vel[:,::].max(axis=0)

print "vref = %r" % vref
print "vborg = %r" % vborg

print "ratio = %r" % (vborg/vref)



#+
#   ARES/HADES/BORG Package -- ./scripts/merge_mpi_restart.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
import os
import h5py as h5
import errno
from ares_tools import rebuild_spliced_h5

def detect_ncpus(path):
  ncpu = 0
  try:
    while True:
      with open("%s_%d" % (path,ncpu), mode= "rb") as f:
        ncpu += 1
  except IOError as e:
    if e.errno != errno.ENOENT:
      raise e
      
  return ncpu
  
def detect_job_to_merge(path):
  
  array_list=[]
  other_list=[]
  group_list=[]
  
  def _handle_item(name, obj):
    if isinstance(obj, h5.Group):
      group_list.append(name)
    if not isinstance(obj, h5.Dataset):
      return

    if len(obj.shape) >= 3:
      array_list.append(name)
    else:
      other_list.append(name)
      
  with h5.File("%s_0" % path, mode="r") as f:  
    f.visititems(_handle_item)

  return array_list,other_list,group_list

def load_nonarray(path, objlist):
  arr = {}
  with h5.File("%s_0" % path, mode="r") as f:  
    for oname in objlist:
      print("Loading %s..." % oname)
      if oname == '/scalars/BORG_version':
        arr[oname] = np.array([f[oname][0]], dtype='S')
      else:
        arr[oname] = f[oname][:]
      
  return arr
  
def load_merged(path):

  ncpu = detect_ncpus(path)
  array_list,nonarray_list,group_list = detect_job_to_merge(path)
  
  array_elts = ['.'.join(e.split('/')) for e in array_list]

  print("Loading spliced arrays")  
  arr = load_nonarray(path, nonarray_list)
  arr2 = rebuild_spliced_h5(path, array_elts, ncpu, verbose=True)
  for k in arr2.keys():
    arr['/'.join(k.split('.'))] = arr2[k]
  
  return arr,group_list
  
  
def save_merged(outpath, omap):
  omap,group_list = omap
  with h5.File(outpath, mode="w") as f:
    for g in group_list:
      if not g in f:
        f.create_group(g)

    for o in omap.keys():
      print("Saving object '%s'" % o)
      f.create_dataset(o, data=omap[o])
      
      


if __name__=="__main__":
  merged_obj = load_merged("./restart.h5")
  
  save_merged("./merged_restart.h5", merged_obj)



#
# Example to run a Altair forward model with PyBORG
#
import borg
import numpy as np


# Setup resolution of the initial mesh
Ng = 64
# Box size in Mpc/h
L = 8000.0

# setup the box
bb = borg.forward.BoxModel()
bb.L = L, L, L
bb.N = Ng, Ng, Ng
bb.xmin = -L/2,-L/2,-L/2

print(bb)

# Initialize some default cosmology
cosmo = borg.cosmo.CosmologicalParameters()

# Fiducial scale factor to express initial conditions
a0 = 0.1

chain = borg.forward.ChainForwardModel(bb)
# Add primordial fluctuations
chain.addModel(borg.forward.models.Primordial(bb, a0))
# Add E&Hu transfer function
chain.addModel(borg.forward.models.EisensteinHu(bb))
# Run an LPT model from a=0.0 to af. The ai=a0 is the scale factor at which the IC are expressed
lpt = borg.forward.models.BorgLpt(bb, bb, ai=a0, af=1.0)
chain.addModel(lpt)

Lz=10000
altair = borg.forward.models.newModel(
    "ALTAIR_AP",
    bb,
    dict(
        corner0_z=-Lz/2,
        corner1_z=-Lz/2,
        corner2_z=-Lz/2,
        L0_z=Lz,
        L1_z=Lz,
        L2_z=Lz,
        N0_z=64,
        N1_z=64,
        N2_z=64,
        is_contrast=True
    ),
)
chain.addModel(altair)

# Set the cosmology
chain.setCosmoParams(cosmo)


# Generate white noise: it has to be scaled by 1/N**(3./2) to be one in Fourier
ic = np.random.randn(Ng, Ng, Ng) / np.sqrt(Ng ** 3)
delta_m = np.zeros(lpt.getOutputBoxModel().N)
delta_m_ap = np.zeros(chain.getOutputBoxModel().N)

# RUN!
chain.forwardModel_v2(ic)
lpt.getDensityFinal(delta_m)
chain.getDensityFinal(delta_m_ap)

# Obtain the particles
pp = np.zeros((lpt.getNumberOfParticles(), 3))
lpt.getParticlePositions(pp)

print(pp)



import numpy as np
from .borg_native import *


class DFT:
    """This implements an MPI DFT transform which is compatible with the other IO involved
    in BORG.

    Note that r2c and c2r are inverse and forward transform respectively. They are already
    properly normalized with physical units.

    TODO: Specify sub-communicators

    Arguments:
      box (BoxModel): Physical box specifications of the transform

    """
    def __init__(self, box):
        # We do an abuse of the API here
        self._box = box
        self._model = forward.models.HadesLinear(box, ai=1.0, af=1.0)
        cpar = cosmo.CosmologicalParameters()
        self._model.setCosmoParams(cpar)
        _, localN0, N1, N2 = self._model.getMPISlice()
        self.Nfft = localN0, N1, N2 // 2 + 1
        self.Nreal = localN0, N1, N2

    def r2c(self, real_array):
        """Execute a real to complex transform

        Arguments:
          real_array (numpy.ndarray): Input real array

        Returns:
          numpy.ndarray: A complex array
        """
        assert real_array.dtype == np.float64
        self._model.forwardModel_v2(real_array)

        out = np.empty(self.Nfft, dtype=np.complex128)
        self._model.getDensityFinal(out)
        return out

    def c2r(self, complex_array):
        """Execute a complex to real transform

        Arguments:
          real_array (numpy.ndarray): Input complex array

        Returns:
          numpy.ndarray: A real array
        """
        assert real_array.dtype == np.complex128
        self._model.forwardModel_v2(complex_array)

        out = np.empty(self.Nreal, dtype=np.float64)
        self._model.getDensityFinal(out)
        return out



#+
#   ARES/HADES/BORG Package -- ./build_tools/gather_sources.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
# ARES TAG: authors_num = 1
# ARES TAG: name(0) = Guilhem Lavaux
# ARES TAG: email(0) = guilhem.lavaux@iap.fr
# ARES TAG: year(0) = 2014-2020
import sys
import shutil
import tempfile
import re
import time
from git import Repo,Tree,Blob,Submodule

distribute_text=""

def build_slice(start, end):
  if end == start:
    return str(start)
  else:
    return str(start)+"-"+str(end)

def line_injection(tag, text, func):
  output = []
  for line in text.split('\n'):
    if line.find(tag) >= 0:
      line = func(line)
    output.append(line)
  return "\n".join(output)

def build_years(years):
  year_list = []
  start = prev_y = years[0]
  for y in years[1:]:
    if y != prev_y+1:
      year_list.append(build_slice(start, prev_y))
      start = y
    prev_y = y
  year_list.append(build_slice(start, prev_y))

  return ", ".join(year_list) #[str(y) for y in years])

class BadFileData(Exception):
  pass

def checked_author(data, i):
  defaults = {
     'name(0)': 'Guilhem Lavaux',
     'email(0)': 'guilhem.lavaux@iap.fr',
     'year(0)': '2014-2020',
     'name(1)': 'Jens Jasche',
     'email(1)': 'jens.jasche@fysik.su.se',
     'year(1)': '2009-2020'
  }
  codes = ['name(%d)' % i, 'email(%d)' % i, 'year(%d)' % i]
  if any(not s in data for s in codes):
    if i <= 1:
      print("   \033[1mWARNING: Using default author data. Please fix file.\033[0m")
    else:
      print("   \033[41mERROR: Need more author data. Please fix file.\033[0m")
      raise BadFileData()

  defaults.update(data)
  return tuple(map(lambda k: defaults[k], codes))

def main_author_handler(line, data, fname):
  num = int(data.get('authors_num', 2))
  output = []
  for i in range(num):
     name, email, year = checked_author(data, i)
     this_line = re.sub('@MAIN_NAME@', name, line)
     this_line = re.sub('@MAIN_EMAIL@', email, this_line)
     output.append( re.sub('@MAIN_YEAR@', year, this_line) )
  return "\n".join(output)

discard_set = set(['Temp'])

def apply_license(license, relimit, filename, authors):
  header = re.sub(r'@FILENAME@', filename, license)
  # Look for the AUTHORS tag in the license template, it has to be support both type of comment.
  m = re.search(
      r'^([@#/<>()\w\-*+ \t\n:.]+)\n([#/()\w\-* \t]*)@AUTHORS@([#/()\w\-* \t]*)\n([@#/()\w\-*+ \n:.;,<>]+)$',
      header, flags=re.MULTILINE)
  init_header,pre_author,post_author,final_header = m.group(1,2,3,4)
  header = init_header + '\n'
  author_list = list( authors.keys())
  author_list.sort()
  for author in author_list:
    if author in discard_set:
      continue
    a_data = authors[author]
    email = a_data['email']
    years = a_data['years']
    years = build_years(years)
    header += pre_author + ("%(name)s <%(email)s> (%(years)s)" % dict(name=author,email=email, years=years)) + post_author + '\n'
  header += final_header

  m = re.search(
      r'^([@#/<>(),\w\-*+ \t\n:.]+)\n([#/()\w\-* \t]*)@DISTRIBUTE@([#/()\w\-* \t]*)\n([@#/()\w\-*+ \n:.;,<>]+)$',
      header, flags=re.MULTILINE)
  if m is None:
     print("We reached an invalid state.")
     print(f"Header is:\n{header}")
     sys.exit(1)
  init_header,pre_distribute,post_distribute,final_header = m.group(1,2,3,4)
  header = f"{init_header}\n"
  for distribute_line in distribute_text.split('\n'):
    header += f"{pre_distribute}{distribute_line}{post_distribute}\n"
  header += final_header

  with open(filename, mode="rt", encoding="UTF-8") as f:
      lines = f.read()

  # Now look for the tag section
  specials = {}
  for a in re.finditer(r"(#|//) ARES TAG:[ \t]*(?P<tag>[\w()]+)[ \t]*=[ \t]*(?P<value>[\w\t \-_\.@]*)", lines):
    b = a.groupdict()
    specials[b['tag']] = b['value']

  header = line_injection('@MAIN_NAME@', header, lambda l: main_author_handler(l, specials, filename))

  lines = re.sub(relimit, lambda x: (("" if x.group(1) is None else x.group(1)) + header), lines)

  with tempfile.NamedTemporaryFile(delete=False,encoding="UTF-8",mode="wt") as tmp_sources:
    tmp_sources.write(lines)

  shutil.move(tmp_sources.name, filename)

def apply_python_license(filename, authors):
  license="""#+
#   ARES/HADES/BORG Package -- @FILENAME@
#   Copyright (C) @MAIN_YEAR@ @MAIN_NAME@ <@MAIN_EMAIL@>
#
#   Additional contributions from:
#      @AUTHORS@
#   @DISTRIBUTE@
#+
"""

  print("Shell/Python/Julia file: %s" % filename)
  relimit=r'^(#!.*\n)?#\+\n(#.*\n)*#\+\n'
  apply_license(license, relimit, filename, authors)


def apply_cpp_license(filename, authors):
  license="""/*+
    ARES/HADES/BORG Package -- @FILENAME@
    Copyright (C) @MAIN_YEAR@ @MAIN_NAME@ <@MAIN_EMAIL@>

    Additional contributions from:
       @AUTHORS@
    @DISTRIBUTE@
+*/
"""
  relimit = r'(?s)^()/\*\+.*\+\*/\n'
  print("C++ file: %s" % filename)
  apply_license(license, relimit, filename, authors)


def patch_author_list(authors):
  patcher={
     'Guilhem Lavaux':'guilhem.lavaux@iap.fr',
     'Jens Jasche':'j.jasche@tum.de'}
  for a in patcher.keys():
    if a in authors:
      data = authors[a]
      data['email'] = patcher[a]

  author_merge(authors, 'MinhMPA', 'Minh Nguyen')
  author_merge(authors, 'Minh MPA', 'Minh Nguyen')
  author_merge(authors, 'flo', 'Florian Führer')
  author_merge(authors, 'LAVAUX Guilhem', 'Guilhem Lavaux')

def author_merge(authors, a_from, a_to):
  if a_from in authors:
    data1 = authors[a_from]
    del authors[a_from]
    if a_to in authors:
      data2 = authors[a_to]
      s = set(data2['years'])
      for y in data1['years']:
        s.add(y)
      s = list(s)
      s.sort()
      data2['years'] = s
    else:
      authors[a_to] = data1

def check_authors(repo, fname):
  authors={}
  author_names={}
  for c,_ in repo.blame('HEAD',fname,w=True,M=True):
    if not c.author.name in authors:
      authors[c.author.name] = set()
    author_names[c.author.name] = c.author.email
    authors[c.author.name].add(time.gmtime(c.authored_date)[0])

  for k in authors.keys():
    authors[k] = list(authors[k])
    authors[k].sort()

  authors = {k:dict(email=author_names[k],years=authors[k]) for k in authors.keys()}
  patch_author_list(authors)
  return authors

def manage_file(repo, fname):
  authors = check_authors(repo, fname)
  if re.match(".*\.(sh|py|pyx|jl)$",fname) != None:
    apply_python_license(fname, authors)
  if re.match('.*\.(tcc|cpp|hpp|h)$', fname) != None:
    apply_cpp_license(fname, authors)

def analyze_tree(repo, prefix, t):
  for entry in t:
    if type(entry) == Submodule:
#      analyze_tree(prefix + "/" + entry.path, Repo(entry.path).tree())
#entry.module())
      print("Seeing a submodule at path " + entry.path)
      continue
    ename = entry.name
    if ename == 'external' or ename == 'cmake_modules':
      continue
    if type(entry) == Tree:
      analyze_tree(repo, prefix + "/" + ename, entry)
    elif type(entry) == Blob:
      fname=prefix+"/"+ename
      manage_file(repo, fname)


if __name__=="__main__":
  repo = Repo(".")
  assert repo.bare == False
  if len(sys.argv) > 1:
     for f in sys.argv[1:]:
        manage_file(repo, f)
  else:
     t = repo.tree()
     analyze_tree(repo, ".", t)



#+
#   ARES/HADES/BORG Package -- ./scripts/ares_tools/read_all_h5.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
import os
import numpy as np
import numexpr as ne
import h5py as h5

def isnotebook():
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False      # Probably standard Python interpreter


try:
    import tqdm

    if isnotebook():
      u_tqdm = tqdm.tqdm_notebook
    else:
      u_tqdm = tqdm.tqdm

    def progress(iterator):
        L = list(iterator)
        return u_tqdm(L)
except:
    def progress(iterator):
        for i, q in enumerate(iterator):
            if ((i % 100) == 0):
                print(i)
            yield q


def default_slicer(x): return x[...]


class Bunch(object):
    def __init__(self, **kwds):
        self.__dict__.update(**kwds)

    def __del__(self):
        if hasattr(self, '_lazy') and self._lazy:
            # close the file/group
            if isinstance(self._group, h5.File):
                print("Closing HDF5 file")
                try:
                    self._group.close()
                except:
                    # Eat all exceptions
                    pass

    def insert(self, aname, data):
        if len(aname) == 1:
            self.__dict__.update({aname[0]: data})
        else:
            if aname[0] in self.__dict__:
                k = self.__dict__[aname[0]]
            else:
                k = Bunch()
                self.__dict__.update({aname[0]: k})
            k.insert(aname[1:], data)


def read_group(g, lazy=False):
    m = {}
    for k in g:
        if hasattr(g[k], 'keys'):
            m[k] = read_group(g[k], lazy=lazy)
        else:
            if lazy:
                m[k] = g[k]
            else:
                m[k] = g[k][:]
    if lazy:
        m['_group'] = g
    m['_lazy'] = lazy

    return Bunch(**m)


def read_attr_group(g, ename, bunch, slicer=default_slicer):
    for a in ename:
        g = g[a]
    if bunch == None:
        bunch = Bunch()

    bunch.insert(ename, slicer(g))
    return bunch


def read_all_h5(fname, lazy=False):
    if lazy:
        f = h5.File(fname, mode="r")
        return read_group(f, lazy=True)
    else:
        with h5.File(fname, mode="r") as f:
            return read_group(f, lazy=False)


def read_attr_h5(fname, egroup, slicer=default_slicer):
    with h5.File(fname, mode="r") as f:
        b = None
        for e in egroup:
            b = read_attr_group(f, e, b, slicer=slicer)
    return b


def grabber(obj, e):
    if len(e) == 1:
        return getattr(obj, e[0])
    else:
        return grabber(getattr(obj, e[0]), e[1:])


def rebuild_spliced_h5(path, element_list, ncpu, verbose=False, flex_cpu=False):
    """Rebuild a set of fields which are spliced across different hdf5 files.
    The use of this function is typically dedicated to the analysis of MPI run.

    Parameters
    ----------
    path         : string
                   base path for the set of hdf5 file. A suffix "_[CPUID]" will be appended.
    element_list : list of string
                   list of elements to rebuild from the set of files
    ncpu         : int
                   number of cpus for the run
    verbose      : boolean
                   if True the code will run with verbose output
    flex_cpu     : boolean
                   if True, ncpu is understood as the maximum number of cpus. If the code does
                   find a file then it stops the rebuilding without failing.

    Returns
    -------
       dictionnary
       each element name is a key, and the value is a numpy array. The arrays are concatenated
       according to their first axis.
    """
    b = [None for e in element_list]
    egrab = [e.split('.') for e in element_list]

    for cpu in range(ncpu):
        fname = path + "_%d" % cpu
        if verbose:
            print("Loading CPU file '%s'" % fname)
        try:
            a = read_attr_h5(fname, egrab)
        except OSError:
            if flex_cpu:
                break
            raise
        for j, e in enumerate(egrab):
            if b[j] is None:
                b[j] = grabber(a, e)
            else:
                b[j] = np.append(b[j], grabber(a, e), axis=0)

    dtype = [(e, t[0].dtype, t[0].shape) for e, t in zip(element_list, b)]
    arr = {}
    for e, q in zip(element_list, b):
        arr[e] = q
    return arr


def chain_iterator(path, start=0, step=1, err_up=0, end=-1, prefix="mcmc", need_id=False):
    import os

    i = start
    while True:
        fname = os.path.join(path, "%s_%d.h5" % (prefix,i,))
        try:
            os.stat(fname)
        except IOError:
            if (i >= err_up):
                return
            else:
                i += step
                continue

        if need_id:
          yield (i,fname)
        else:
          yield fname
        i += step
        if end > 0 and i > end:
          break


def read_chain_h5(path, element_list, start=0, step=1, err_up=0, slicer=default_slicer, prefix="mcmc", flexible=True):
    """
    read_chain_h5(path,element_list,start=0,step=1,err_up=0,slicer=default_slicer)

    Arguments:
      * path: path where you have the chain (mcmc_* files)
      * element_list: list of strings where the MCMC objects are stored. For example, you
                      have scalars.galaxy_nmean_0 for the nbar parameter of the first catalog.
      * start: the first element of the chain to consider
      * step: if you want to thin the chain by "step"
      * err_up: whether to accept I/O errors when opening files up to the specified MCMC id
      * slicer: a lambda function that can only take a subset of the array of the
      specified MCMC object. For example, it can be lambda d: d[:,:,64], to indicate only
       the plane 64 in the 3d grid that is being loaded. [64,...]

    Returns:
      * a columned numpy array "a". You have one column for each element_id of the
      element_list.  You can access one of the column like this:
      a["scalars.galaxy_nmean_0"]

  """
    i = start
    b = [[] for e in element_list]
    egrab = [e.split('.') for e in element_list]
    for fname in progress(chain_iterator(path, start=start, step=step, err_up=err_up, prefix=prefix)):
        try:
          a = read_attr_h5(fname, egrab, slicer=slicer)
        except OSError:
          if not flexible:
              raise
          else:
              break
        for j, e in enumerate(egrab):
            b[j].append(grabber(a, e))

    dtype = [(e, t[0].dtype, t[0].shape) for e, t in zip(element_list, b)]

    arr = np.empty(len(b[0]), dtype=dtype)
    for e, q in zip(element_list, b):
        arr[e] = q
    return arr


def chain_compute_xcor(path, start=0, Nbins=100):
    import cosmotool as ct
    import numexpr as ne

    i = 0

    with h5.File(os.path.join(path, "restart.h5_0"), mode="r") as f:
        L0 = f['scalars']['L0'][:]
        L1 = f['scalars']['L1'][:]
        L2 = f['scalars']['L2'][:]
        N0 = int(f['scalars']['N0'][:])
        N1 = int(f['scalars']['N1'][:])
        N2 = int(f['scalars']['N2'][:])

    ix = np.fft.fftfreq(N0, d=1.0/L0)[:, None, None]
    iy = np.fft.fftfreq(N1, d=1.0/L1)[None, :, None]
    iz = np.fft.fftfreq(N2, d=1.0/L2)[None, None, :]
    r2 = ne.evaluate('sqrt(ix**2+iy**2+iz**2)')
    rmax = r2.max()
    ir = (r2 * Nbins / rmax).astype(np.int32).ravel()
    xi = []
    W = np.bincount(ir, minlength=Nbins)

    fft = ct.CubeFT(L0, N0)

    while True:
        try:
            if i % 10 == 0:
                print(i)
            fname = os.path.join(path, "mcmc_%d.h5" % (i+start))
            with h5.File(fname, mode="r") as f:
                fft.density = f["scalars"]["s_field"][:]

            fft.rfft()
            ne.evaluate("complex(real(d)**2 + imag(d)**2, 0)",
                        local_dict={'d': fft.dhat}, out=fft.dhat, casting='unsafe')
            fft.irfft()

            xi.append(np.bincount(
                ir, weights=fft.density.ravel(), minlength=Nbins))
            i += 1

        except Exception as e:
            print(repr(e))
            break

    xi = np.array(xi) / W

    r = np.arange(Nbins) * rmax / Nbins
    return r, xi


def explore_chain(path, start, end=-1, step=1, quiet=True):
    """
    Arguments:
      * path
      * start
      * end
      * step

    Returns:
      * iterator with hdf5 object. Example:
           for i in explore_chain(".", 0):
              mean = i['/scalars/galaxy_nmean_0'][0]
              # Then do stuff with "mean"
  """
    n = int(start)
    nmax = int(end)
    step = int(step)
    k = 0

    while (nmax == -1) or (n < nmax):
        p = path + "/mcmc_%d.h5" % n
        if not quiet and (k % 100) == 0:
            print("%d" % n)
        try:
            f = h5.File(p, mode="r")
        except Exception as e:
            print(e)
            break

        try:
            yield n, f['scalars']
        finally:
            f.close()

        n += step
        k += 1


def build_power_histogram(path, start=0, step=1, Nhisto=100, quiet=True, logP=True, Prange=(0.1, 1e5)):
    """
    Use the scalars.powerspectrum mcmc element to build the PDF of the posterior
    powerspectrum

    Arguments:
       * path
       * start
       * step
       * Nhisto: number of bins for each k mode of the P(k) histogram
       * quiet:
       * logP: whether you want to use a log scale for plotting
       * Prange: a tuple for giving the entire P range to represent

     Returns:
       * a tuple: t=(kmodes, Pgrid, Pk_pdf_values) which is directly usable in pcolormesh
         like pcolormesh(*t)
  """
    # print path+ "/restart.h5_0"
    for _, scalars in explore_chain(path, start, end=start+1, step=1, quiet=quiet):
        Nk = scalars["powerspectrum"].size
    with h5.File(path + "/restart.h5_0", mode="r") as f:
        k_mode = f["/scalars/k_modes"][:]

    Phisto = np.zeros((Nk, Nhisto), dtype=np.int)
    if logP:
        logPmin = np.log10(Prange[0])
        DeltaLog = np.log10(Prange[1]/Prange[0])/(Nhisto)

        def transform(P): return ((np.log10(P)-logPmin)/DeltaLog)
    else:
        Pmin = Prange[0]
        DeltaP = (Prange[1]-Prange[0])/(Nhisto)

        def transform(P): return (P-Pmin)/DeltaP

    for n, scalars in explore_chain(path, start, step=step, quiet=quiet):
        P = scalars["powerspectrum"][:]
        iP = np.floor(transform(P))
        ik = np.where((np.isnan(iP) == False)*(iP >= 0)*(iP < Nhisto))
        iP = iP[ik].astype(np.int)
        for i, j in zip(ik[0], iP):
            Phisto[i, j] = Phisto[i, j]+1

    k_mode = k_mode[:, None].repeat(Nhisto, axis=1)
    if logP:
        Pg = 10**((np.arange(Nhisto)+0.5)*DeltaLog + logPmin)
    else:
        Pg = ((np.arange(Nhisto)+0.5)*DeltaP + Pmin)

    Pg = Pg[None, :].repeat(Nk, axis=0)
    return k_mode, Pg, Phisto


def read_chain_complex_avg_dev(path, op, start=0, end=-1, do_dev=False, step=1, slicer=default_slicer, prefix="mcmc"):
    """
    Compute mean and standard deviation of the given element_list

    Arguments:
      * path
      * op:
      * element_list
      * start
      * do_dev: boolean for computing the standard deviation (or not)
      * slicer:


     Returns:
      * a columned numpy array. Each column has a name that corresponds to an element with an additional dimension. For example, a['scalars.galaxy_nmean_0'][0] -> mean,
      a['scalars.galaxy_nmean_0'][1] is the standard deviation.
  """
    i = 0
    b = None
    bd = None
    try:
      for fname in progress(chain_iterator(path, start=start, step=step, end=end, prefix=prefix)):
        with h5.File(fname, mode="r") as ff:
            if b is None:
                b = op(ff)
                if do_dev:
                    bd = np.zeros(b.shape)
            else:
                data = op(ff)
                ne.evaluate('r*a+k*c',
                            dict(k=1/float(i+1),
                                 r=(float(i)/float(i+1)),
                                 a=b,
                                 c=data),
                            out=b)
            if do_dev and i > 1:
                ne.evaluate('k*(xn-mun)**2 + f*bdn',
                            dict(k=1/float(i),
                                 f=float(i)/float(i+1),
                                 xn=data,
                                 mun=b,
                                 bdn=bd),
                            out=bd)
            i += 1

    except OSError:
      pass

    bd = np.sqrt(bd)
    return b, bd


def read_chain_avg_dev(path, element_list, start=0, end=-1, do_dev=False, operator=lambda x: x, step=1, slicer=default_slicer, prefix="mcmc", err_up=0):
    """
    Compute mean and standard deviation of the given element_list

    Arguments:
      * path
      * element_list
      * start
      * do_dev: boolean for computing the standard deviation (or not)
      * operator: applies the operator on all the elements before computing the mean and
        standard deviation.
      * slicer:


     Returns:
      * a columned numpy array. Each column has a name that corresponds to an element with an additional dimension. For example, a['scalars.galaxy_nmean_0'][0] -> mean,
      a['scalars.galaxy_nmean_0'][1] is the standard deviation.
  """
    i = 0
    b = [None for e in element_list]
    bd = [None for e in element_list]
    egrab = [e.split('.') for e in element_list]
    for fname in progress(chain_iterator(path, start=start, step=step, end=end, err_up=err_up, prefix=prefix)):
            a = read_attr_h5(fname, egrab, slicer=slicer)
            for j, e in enumerate(egrab):
                if b[j] is None:
                    b[j] = operator(grabber(a, e))
                    if do_dev:
                        bd[j] = np.zeros(b[j].shape)
                else:
                    data = operator(grabber(a, e))
                    ne.evaluate('r*a+k*c',
                                dict(k=1/float(i+1),
                                     r=(float(i)/float(i+1)),
                                     a=b[j],
                                     c=data),
                                out=b[j])
                    if do_dev and i > 0:
                        ne.evaluate('k*(xn-mun)**2 + f*bdn',
                                    dict(k=1/float(i),
                                         f=float(i)/float(i+1),
                                         xn=data,
                                         mun=b[j],
                                         bdn=bd[j]),
                                    out=bd[j])
                i+=1

    dtype = [(e, t.dtype, t.shape) for e, t in zip(element_list, b)]

    arr = np.empty(2 if do_dev else 1, dtype=dtype)
    for e, q, q2 in zip(element_list, b, bd):
        arr[e][0] = q
        if do_dev:
            arr[e][1] = np.sqrt(q2)
    return arr



import stat
import os
import sys
import shutil
from sysconfig import get_config_var
from distutils.command.install_data import install_data
import pathlib
from setuptools import find_packages, setup, Extension
from setuptools.command.build_ext import build_ext
from setuptools.command.install_lib import install_lib
from setuptools.command.install_scripts import install_scripts
import struct

BITS = struct.calcsize("P") * 8
PACKAGE_NAME = "aquila_borg"

class CMakeExtension(Extension):
    """
    An extension to run the cmake build

    This simply overrides the base extension class so that setuptools
    doesn't try to build your sources for you
    """

    def __init__(self, name, sources=[]):

        super().__init__(name = name, sources = sources)

        self.SOURCE_DIR = str(pathlib.Path().absolute())

class InstallCMakeLibsData(install_data):
    """
    Just a wrapper to get the install data into the egg-info

    Listing the installed files in the egg-info guarantees that
    all of the package files will be uninstalled when the user
    uninstalls your package through pip
    """

    def run(self):
        """
        Outfiles are the libraries that were built using cmake
        """

        # There seems to be no other way to do this; I tried listing the
        # libraries during the execution of the InstallCMakeLibs.run() but
        # setuptools never tracked them, seems like setuptools wants to
        # track the libraries through package data more than anything...
        # help would be appriciated

        self.outfiles = self.distribution.data_files

class InstallCMakeLibs(install_lib):
    """
    Get the libraries from the parent distribution, use those as the outfiles

    Skip building anything; everything is already built, forward libraries to
    the installation step
    """

    def run(self):
        """
        Copy libraries from the bin directory and place them as appropriate
        """

        self.announce("Moving library files", level=3)
#        print(self.build_temp)
#
#        self.distribution.bin_dir = CosmoTool_extension.bin_dir
#
#        # We have already built the libraries in the previous build_ext step
#
#        self.skip_build = True
#
#        # Depending on the files that are generated from your cmake
#        # build chain, you may need to change the below code, such that
#        # your files are moved to the appropriate location when the installation
#        # is run
#
#        libs = [os.path.join(bin_dir, _lib) for _lib in
#                os.listdir(bin_dir) if
#                os.path.isfile(os.path.join(bin_dir, _lib)) and
#                os.path.splitext(_lib)[1] in [".dll", ".so"]
#                and not (_lib.startswith("python") or _lib.startswith(PACKAGE_NAME))]
#
#        for lib in libs:
#
#            shutil.move(lib, os.path.join(self.build_dir,
#                                          os.path.basename(lib)))
#
#        # Mark the libs for installation, adding them to
#        # distribution.data_files seems to ensure that setuptools' record
#        # writer appends them to installed-files.txt in the package's egg-info
#        #
#        # Also tried adding the libraries to the distribution.libraries list,
#        # but that never seemed to add them to the installed-files.txt in the
#        # egg-info, and the online recommendation seems to be adding libraries
#        # into eager_resources in the call to setup(), which I think puts them
#        # in data_files anyways.
#        #
#        # What is the best way?
#
#        # These are the additional installation files that should be
#        # included in the package, but are resultant of the cmake build
#        # step; depending on the files that are generated from your cmake
#        # build chain, you may need to modify the below code
#
#        self.distribution.data_files = [os.path.join(self.install_dir,
#                                                     os.path.basename(lib))
#                                        for lib in libs]
#        print(self.distribution.data_files)
#
#        # Must be forced to run after adding the libs to data_files

        self.distribution.run_command("install_data")

        super().run()

class BuildCMakeExt(build_ext):
    """
    Builds using cmake instead of the python setuptools implicit build
    """

    def run(self):
        """
        Perform build_cmake before doing the 'normal' stuff
        """

        for extension in self.extensions:

            if extension.name == 'borg':
                self.package = 'borg'

                self.build_cmake(extension)

        #super().run()

    def build_cmake(self, extension: Extension):
        """
        The steps required to build the extension
        """

        self.announce("Preparing the build environment", level=3)

        package_dir = os.path.abspath(os.path.join(self.build_lib, 'borg'))


        extension.build_dir = pathlib.Path(self.build_temp)
        install_dir = os.path.abspath(self.build_lib)
        #extension.bin_dir = str(pathlib.Path(os.path.join(extension.build_dir, 'private_install')).absolute())
        SOURCE_DIR = extension.SOURCE_DIR
        build_dir = extension.build_dir

        extension_path = pathlib.Path(self.get_ext_fullpath(extension.name))

        os.makedirs(build_dir, exist_ok=True)
        os.makedirs(extension_path.parent.absolute(), exist_ok=True)

        # Now that the necessary directories are created, build

        c_compiler=os.environ.get('CC', get_config_var("CC"))
        cxx_compiler=os.environ.get('CXX', get_config_var("CXX"))

        os.environ['MACOSX_DEPLOYMENT_TARGET'] = '10.14'
        install_dir = os.path.abspath(self.build_lib)

        self.announce("Configuring cmake project", level=3)

        self.spawn(['bash', 'build.sh', '--build-dir', self.build_temp,
                    f"--python={sys.executable}", "--c-compiler", c_compiler,
                    "--cxx-compiler", cxx_compiler, '--purge',
                    '--install-system-python',
                    '--extra-flags', f"-DPYTHON_SITE_PACKAGES={install_dir} -DINSTALL_COMPATIBILITY_PYBORG=OFF -DINSTALL_SHARED=ON -DDATA_INSTALL_DIR={install_dir}/aquila_borg/_data"
                    ])

        self.announce("Building binaries", level=3)

        self.spawn(["cmake", "--build", self.build_temp, "--target", "install",
                      "--config", "Release"])

        # Build finished, now copy the files into the copy directory
        # The copy directory is the parent directory of the extension (.pyd)

pyborg_extension = CMakeExtension(name="borg")

version_py = os.path.join(os.path.dirname(__file__), 'VERSION.txt')
with open(version_py, "r") as gh:
  version = gh.read().strip()

setup(name='aquila_borg',
      version=version,
      packages=["aquila_borg"],
      package_dir={'aquila_borg': 'extra/python/python/aquila_borg'},
      install_requires=['numpy','deprecated'],
      setup_requires=['numpy'],
      ext_modules=[pyborg_extension],
      description='ARES/BORG engine packaged in a Python module',
      long_description=open("./README.rst", 'r').read(),
      long_description_content_type="text/x-rst",
      keywords="cosmology, inference, simulation, statistics",
      classifiers=["Intended Audience :: Developers",
                   "Natural Language :: English",
                   "License :: OSI Approved :: "
                   "GNU Lesser General Public License v3 (LGPLv3)",
                   "License :: OSI Approved :: "
                   "CEA CNRS Inria Logiciel Libre License, version 2.1 (CeCILL-2.1)",
                   "Programming Language :: C",
                   "Programming Language :: C++",
                   "Programming Language :: Python",
                   "Programming Language :: Python :: 3",
                   "Programming Language :: Python :: Implementation :: CPython"],
      license_files=('License_CeCILL_V2.1.txt','License_GPL-3.0.txt'),
      include_package_data=True,
      cmdclass={
          'build_ext': BuildCMakeExt,
          'install_data': InstallCMakeLibsData,
          'install_lib': InstallCMakeLibs,
          },
      python_requires='>=3.6'
)



#+
#   ARES/HADES/BORG Package -- ./extra/python/python/aquila_borg/borg_native.py
#   Copyright (C) 2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
try:
    from borg_embed import *
    EMBEDDED = True
except ModuleNotFoundError as e:
    # If we do not detect the embedded module we add some
    # symbols to account for the missing decorator functions
    # That way the python code are working both embedded or
    # in standalone python.
    from ._borg import *

    def _warning_function(f):
        print("warning! not embedded borg")
        return f

    registerLikelihoodBuilder = _warning_function
    registerGravityBuilder = _warning_function
    registerSamplerBuilder = _warning_function
    EMBEDDED = False
    del _warning_function

__all__ = [
    "cosmo", "forward", "bias", "likelihood", "samplers", "Console", "console",
    "memoryReport", "registerLikelihoodBuilder", "registerGravityBuilder",
    "registerSamplerBuilder", "EMBEDDED"
]


# ARES TAG: authors_num = 1
# ARES TAG: name(0) = Guilhem Lavaux
# ARES TAG: year(0) = 2020
# ARES TAG: email(0) = guilhem.lavaux@iap.fr



#+
#   ARES/HADES/BORG Package -- ./scripts/notebooks/src/curl.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
import numpy as np

def curl(vect,dx=1,dy=1,dz=1):
    "return the curl of a n-D field"
    [P_dy,P_dz]=np.gradient(vect[0],axis=[1,2])
    [Q_dx,Q_dz]=np.gradient(vect[1],axis=[0,2])
    [R_dx,R_dy]=np.gradient(vect[2],axis=[0,1])
    
    curl=np.array([R_dy-Q_dz,P_dz-R_dx,Q_dx-P_dy])
    
    return curl

def div(vect,dx=1,dy=1,dz=1):
    "return the divergence of a n-D field"
    return np.sum(np.gradient(vect),axis=0)
             



#+
#   ARES/HADES/BORG Package -- ./scripts/misc/check_bias.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
import read_all_h5
from pylab import *

n=[]
b0=[]
b1=[]
b2=[]
b3=[]
b4=[]
b5=[]
b6=[]
b7=[]
b8=[]
b9=[]
b10=[]
b11=[]
b12=[]
b13=[]
b14=[]
b15=[]


rho_g0=[]
rho_g1=[]
rho_g2=[]
rho_g3=[]
rho_g4=[]
rho_g5=[]
rho_g6=[]
rho_g7=[]
rho_g8=[]
rho_g9=[]
rho_g10=[]
rho_g11=[]
rho_g12=[]
rho_g13=[]
rho_g14=[]
rho_g15=[]


eps_g0=[]
eps_g1=[]
eps_g2=[]
eps_g3=[]
eps_g4=[]
eps_g5=[]
eps_g6=[]
eps_g7=[]
eps_g8=[]
eps_g9=[]
eps_g10=[]
eps_g11=[]
eps_g12=[]
eps_g13=[]
eps_g14=[]
eps_g15=[]


n0=[]
n1=[]
n2=[]
n3=[]
n4=[]
n5=[]
n6=[]
n7=[]
n8=[]
n9=[]
n10=[]
n11=[]
n12=[]
n13=[]
n14=[]
n15=[]


accept=[]

i=0
#while True:



for l in range(0,440,1):
  a = \
   read_all_h5.read_all_h5("mcmc_%d.h5" % l)
  try:
    n0.append(a.scalars.galaxy_nmean_0)
    n1.append(a.scalars.galaxy_nmean_1)
    n2.append(a.scalars.galaxy_nmean_2)
    n3.append(a.scalars.galaxy_nmean_3)
    n4.append(a.scalars.galaxy_nmean_4)
    n5.append(a.scalars.galaxy_nmean_5)
    n6.append(a.scalars.galaxy_nmean_6)
    n7.append(a.scalars.galaxy_nmean_7)
    n8.append(a.scalars.galaxy_nmean_8)
    n9.append(a.scalars.galaxy_nmean_9)
    n10.append(a.scalars.galaxy_nmean_10)
    n11.append(a.scalars.galaxy_nmean_11)
    n12.append(a.scalars.galaxy_nmean_12)
    n13.append(a.scalars.galaxy_nmean_13)
    n14.append(a.scalars.galaxy_nmean_14)
    n15.append(a.scalars.galaxy_nmean_15)
    
    
    b0.append(a.scalars.galaxy_bias_0)
    b1.append(a.scalars.galaxy_bias_1)
    b2.append(a.scalars.galaxy_bias_2)
    b3.append(a.scalars.galaxy_bias_3)
    b4.append(a.scalars.galaxy_bias_4)
    b5.append(a.scalars.galaxy_bias_5)
    b6.append(a.scalars.galaxy_bias_6)
    b7.append(a.scalars.galaxy_bias_7)
    b8.append(a.scalars.galaxy_bias_8)
    b9.append(a.scalars.galaxy_bias_9)
    b10.append(a.scalars.galaxy_bias_10)
    b11.append(a.scalars.galaxy_bias_11)
    b12.append(a.scalars.galaxy_bias_12)
    b13.append(a.scalars.galaxy_bias_13)
    b14.append(a.scalars.galaxy_bias_14)
    b15.append(a.scalars.galaxy_bias_15)
    
    rho_g0.append(a.scalars.galaxy_rho_g_0)
    rho_g1.append(a.scalars.galaxy_rho_g_1)
    rho_g2.append(a.scalars.galaxy_rho_g_2)
    rho_g3.append(a.scalars.galaxy_rho_g_3)
    rho_g4.append(a.scalars.galaxy_rho_g_4)
    rho_g5.append(a.scalars.galaxy_rho_g_5)
    rho_g6.append(a.scalars.galaxy_rho_g_6)
    rho_g7.append(a.scalars.galaxy_rho_g_7)
    rho_g8.append(a.scalars.galaxy_rho_g_8)
    rho_g9.append(a.scalars.galaxy_rho_g_9)
    rho_g10.append(a.scalars.galaxy_rho_g_10)
    rho_g11.append(a.scalars.galaxy_rho_g_11)
    rho_g12.append(a.scalars.galaxy_rho_g_12)
    rho_g13.append(a.scalars.galaxy_rho_g_13)
    rho_g14.append(a.scalars.galaxy_rho_g_14)
    rho_g15.append(a.scalars.galaxy_rho_g_15)
    
    eps_g0.append(a.scalars.galaxy_eps_g_0)
    eps_g1.append(a.scalars.galaxy_eps_g_1)
    eps_g2.append(a.scalars.galaxy_eps_g_2)
    eps_g3.append(a.scalars.galaxy_eps_g_3)
    eps_g4.append(a.scalars.galaxy_eps_g_4)
    eps_g5.append(a.scalars.galaxy_eps_g_5)
    eps_g6.append(a.scalars.galaxy_eps_g_6)
    eps_g7.append(a.scalars.galaxy_eps_g_7)
    eps_g8.append(a.scalars.galaxy_eps_g_8)
    eps_g9.append(a.scalars.galaxy_eps_g_9)
    eps_g10.append(a.scalars.galaxy_eps_g_10)
    eps_g11.append(a.scalars.galaxy_eps_g_11)
    eps_g12.append(a.scalars.galaxy_eps_g_12)
    eps_g13.append(a.scalars.galaxy_eps_g_13)
    eps_g14.append(a.scalars.galaxy_eps_g_14)
    eps_g15.append(a.scalars.galaxy_eps_g_14)
    
    accept.append(a.scalars.hades_accept_count)
    print l
  except AttributeError:
    break
  i += 1

rate =np.cumsum(np.array(accept))
norm =np.cumsum(np.ones(len(accept)))

plt.plot(rate/norm)
plt.show()

print 

plt.plot(b0,label=str(0))
plt.plot(b1,label=str(1))
plt.plot(b2,label=str(2))
plt.plot(b3,label=str(3))
plt.plot(b4,label=str(4))
plt.plot(b5,label=str(5))
plt.plot(b6,label=str(6))
plt.plot(b7,label=str(7))
plt.plot(b8,label=str(0))
plt.plot(b9,label=str(1))
plt.plot(b10,label=str(2))
plt.plot(b11,label=str(3))
plt.plot(b12,label=str(4))
plt.plot(b13,label=str(5))
plt.plot(b14,label=str(6))
plt.plot(b15,label=str(7))

legend()
plt.savefig('check_bias.png')
plt.show()


plt.plot(np.log10(rho_g0),label=str(0))
plt.plot(np.log10(rho_g1),label=str(1))
plt.plot(np.log10(rho_g2),label=str(2))
plt.plot(np.log10(rho_g3),label=str(3))
plt.plot(np.log10(rho_g4),label=str(4))
plt.plot(np.log10(rho_g5),label=str(5))
plt.plot(np.log10(rho_g6),label=str(6))
plt.plot(np.log10(rho_g7),label=str(7))
plt.plot(np.log10(rho_g8),label=str(0))
plt.plot(np.log10(rho_g9),label=str(1))
plt.plot(np.log10(rho_g10),label=str(2))
plt.plot(np.log10(rho_g11),label=str(3))
plt.plot(np.log10(rho_g12),label=str(4))
plt.plot(np.log10(rho_g13),label=str(5))
plt.plot(np.log10(rho_g14),label=str(6))
plt.plot(np.log10(rho_g15),label=str(7))
legend()
plt.savefig('check_rho_g.png')
plt.show()



x=np.arange(600)*0.04+1e-12

y0=n0[-1]*x**b0[-1]*np.exp(-rho_g0[-1]*x**(-eps_g0[-1]))
y1=n1[-1]*x**b1[-1]*np.exp(-rho_g1[-1]*x**(-eps_g1[-1]))
y2=n2[-1]*x**b2[-1]*np.exp(-rho_g2[-1]*x**(-eps_g2[-1]))
y3=n3[-1]*x**b3[-1]*np.exp(-rho_g3[-1]*x**(-eps_g3[-1]))
y4=n4[-1]*x**b4[-1]*np.exp(-rho_g4[-1]*x**(-eps_g4[-1]))
y5=n5[-1]*x**b5[-1]*np.exp(-rho_g5[-1]*x**(-eps_g5[-1]))
y6=n6[-1]*x**b6[-1]*np.exp(-rho_g6[-1]*x**(-eps_g6[-1]))
y7=n7[-1]*x**b7[-1]*np.exp(-rho_g7[-1]*x**(-eps_g7[-1]))
y8=n8[-1]*x**b8[-1]*np.exp(-rho_g8[-1]*x**(-eps_g8[-1]))
y9=n9[-1]*x**b9[-1]*np.exp(-rho_g9[-1]*x**(-eps_g9[-1]))
y10=n10[-1]*x**b10[-1]*np.exp(-rho_g10[-1]*x**(-eps_g10[-1]))
y11=n11[-1]*x**b11[-1]*np.exp(-rho_g11[-1]*x**(-eps_g11[-1]))
y12=n12[-1]*x**b12[-1]*np.exp(-rho_g12[-1]*x**(-eps_g12[-1]))
y13=n13[-1]*x**b13[-1]*np.exp(-rho_g13[-1]*x**(-eps_g13[-1]))
y14=n14[-1]*x**b14[-1]*np.exp(-rho_g14[-1]*x**(-eps_g14[-1]))
y15=n15[-1]*x**b15[-1]*np.exp(-rho_g15[-1]*x**(-eps_g15[-1]))


plt.plot(x,np.log(y0),label=str(0),color='red')
plt.plot(x,np.log(y1),label=str(1),color='blue')
plt.plot(x,np.log(y2),label=str(2),color='green')
plt.plot(x,np.log(y3),label=str(3),color='orange')
plt.plot(x,np.log(y4),label=str(4),color='yellow')
plt.plot(x,np.log(y5),label=str(5),color='black')
plt.plot(x,np.log(y6),label=str(6),color='gray')
plt.plot(x,np.log(y7),label=str(7),color='magenta')
plt.plot(x,np.log(y8),label=str(0),color='red')
plt.plot(x,np.log(y9),label=str(1),color='blue')
plt.plot(x,np.log(y10),label=str(2),color='green')
plt.plot(x,np.log(y11),label=str(3),color='orange')
plt.plot(x,np.log(y12),label=str(4),color='yellow')
plt.plot(x,np.log(y13),label=str(5),color='black')
plt.plot(x,np.log(y14),label=str(6),color='gray')
plt.plot(x,np.log(y15),label=str(7),color='magenta')


plt.plot(x,np.log(x),label=str(99))
plt.ylim([-8,5])
legend(loc='lower right', shadow=True)
#plt.show()
gcf().savefig("check_bias.png")











#+
#   ARES/HADES/BORG Package -- ./scripts/check_likelihood_trace.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
import read_all_h5
import h5py as h5
from pylab import *
from analysis.analysis import *

#This routine calculates the log-likelihood for a given catalog
def psi_lh(delta,Nobs,selection,nmean,bias,rho_g,eps_g):
    print 'calculating log likelihood...'
    #only calculate for values inside observed domain
    foo=np.where(selection>0.)
    delta=delta[foo]
    rho=1.+delta+1e-12
    Nobs=Nobs[foo]
    selection=selection[foo]
        
    lamb = selection*nmean*(rho**bias)*np.exp(-rho_g*(rho**(-eps_g)))
    aux= lamb - Nobs*(np.log(selection*nmean)+bias*np.log(rho)-rho_g*(rho**(-eps_g)))
    print 'done!'    
    return np.sum(aux)

chain_path="/scratch/jasche/panphasia_run_pm/"

ares=analysis(chain_path=chain_path,LSS_framework='BORG')
ncat=ares.get_ncat()
#load data and masks
mask=[]
nobs=[]
for i in range(ncat):
    print 'load data of catalog Nr.:',i,'...'
    mask.append(ares.get_mask_spliced(i,ncpu=32))
    nobs.append(ares.get_data_spliced(i,ncpu=32))
    print 'done!'  

#open sample
hh=[]
xx=[]

for l in range(822,1350,1):
    #set log likelihood to zero
    H=0
    #open file and get data for smaple
    with h5.File(chain_path + 'mcmc_'+str(l)+'.h5', mode="r") as f:

        delta = f['scalars']['BORG_final_density'][:]
        haux=np.zeros(ncat+1)
        for i in range(ncat):
            nmean   = f['scalars']['galaxy_nmean_' +str(i)][:]
            bias    = f['scalars']['galaxy_bias_' +str(i)][:]
            rho_g   = f['scalars']['galaxy_rho_g_' +str(i)][:]
            eps_g   = f['scalars']['galaxy_eps_g_' +str(i)][:] 
            
            print nmean,bias,rho_g,eps_g
            
            haux[i] = psi_lh(delta,nobs[i],mask[i] ,nmean,bias,rho_g,eps_g)      
            haux[ncat]+=haux[i]
            print nmean,bias,rho_g,eps_g,haux[i]
        
        hh.append(haux)
        xx.append(l)
        
        hha=np.array(hh)
        xxa=np.array(xx)
        print l
        
        np.savez('lh_trace_pm',hh=hha,xx=xxa)
        
plt.plot(xxa,hha)
plt.show()
            



#+
#   ARES/HADES/BORG Package -- ./scripts/ini_generator/template_sdss_main.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
CONFIG=dict(
  absmag_min=-23.,absmag_max = -17.,
  num_subcat=6,catalog='sdss_ares_cat.txt',
  mask='SDSSDR7MASK_4096.fits', ref_subcat=0,
  cutter='magnitude'
  )

radial_selection = 'schechter'
schechter_mstar = -20.44
schechter_alpha = -1.05
schechter_sampling_rate = 2000
schechter_dmax = 1000
bias = 1
nmean = 1
galaxy_bright_apparent_magnitude_cut = 13.5
galaxy_faint_apparent_magnitude_cut = 17.6



#+
#   ARES/HADES/BORG Package -- ./build_tools/gen_splash.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
import sys
import re

prev_line=None
with open(sys.argv[1], mode="r") as f_in, open(sys.argv[2], mode="w") as f_out:
  for line in f_in:
    if prev_line is not None:
      f_out.write('"' + prev_line + '",\n')
    line = re.sub(r'\\', r'\\\\', line)
    line = re.sub(r'"', r'\"', line)
    prev_line = line[:-1]
  f_out.write('"' + prev_line + '"\n')
